/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.engine3d.entities.base;

import com.jogamp.newt.event.InputEvent;
import com.jogamp.opengl.GL3bc;
import java.io.IOException;
import org.joml.Matrix4f;
import org.joml.Vector3f;
import fr.amap.lidar.amapvox.engine3d.util.InputStatus;
import fr.amap.lidar.amapvox.engine3d.entities.extended.Camera;
import fr.amap.lidar.amapvox.engine3d.entities.extended.Renderer;
import fr.amap.lidar.amapvox.engine3d.resources.materials.SkyboxMaterial;
import fr.amap.lidar.amapvox.engine3d.resources.meshes.DepthFunction;
import fr.amap.lidar.amapvox.engine3d.resources.meshes.Geometry;
import fr.amap.lidar.amapvox.engine3d.resources.meshes.Mesh;
import fr.amap.lidar.amapvox.engine3d.resources.programs.Program;
import fr.amap.lidar.amapvox.engine3d.resources.programs.SkyboxProgram;
import fr.amap.lidar.amapvox.engine3d.resources.textures.TextureCubemap;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.CopyOnWriteArrayList;
import net.jcip.annotations.GuardedBy;
import net.jcip.annotations.ThreadSafe;

/**
 * A parent object that can be used to regroup many entities and make them work.
 */
@ThreadSafe
public abstract class Scene extends Behaviour {

    /**
     * A list of cameras.
     */
    private final CopyOnWriteArrayList<Camera> cameras;

    /**
     * A list of mesh renderers.
     */
    private final CopyOnWriteArrayList<Renderer> renderers;

    /**
     * A list of programs.
     */
    private final ConcurrentMap<String, Program> programs;

    /**
     * A list of other entities.
     */
    private final CopyOnWriteArrayList<Entity> entities;

    /**
     * The background environment skybox program.
     */
    private final SkyboxProgram skyboxProgram;

    /**
     * Cubemap used to draw the skybox.
     */
    @GuardedBy("this")
    private TextureCubemap skyboxCubemap;

    /**
     * The mesh used to draw the skybox.
     */
    private final Mesh skyboxMesh;

    /**
     * If somethings in the scene has been modified and must be draw.
     */
    @GuardedBy("this")
    protected boolean somethingsModified;

    /**
     * If the scene has been drawed twice.
     */
    @GuardedBy("this")
    private int drawCount;

    /**
     * Create and empty scene.
     */
    public Scene() {
        this.cameras = new CopyOnWriteArrayList();
        this.renderers = new CopyOnWriteArrayList();
        this.programs = new ConcurrentHashMap();
        this.entities = new CopyOnWriteArrayList();
        this.skyboxProgram = new SkyboxProgram();
        this.skyboxCubemap = null;
        this.skyboxMesh = new Mesh("SkyboxMesh");
        try {
            this.skyboxMesh.setGeometry(Geometry.loadFromOBJ("/meshes/skybox.obj"));
        } catch (IOException ex) {
            ex.printStackTrace();
        }
        this.skyboxMesh.setDepthFunction(DepthFunction.LEQUAL);
        this.skyboxProgram.setMesh(skyboxMesh);
        this.somethingsModified = true;
        initResources();
        initEntities();
    }

    /**
     * Add an entity to the scene.
     * @param entity
     */
    public final void addEntity(final Entity entity) {
        entity.setScene(this);
        entities.add(entity);

        if (entity.hasComponent(Camera.class)) {
            cameras.add((Camera) entity.getComponent(Camera.class));
        } else if (entity.hasComponent(Renderer.class)) {
            final Renderer renderer = entity.getComponent(Renderer.class);
            renderers.add(renderer);

            /* Create a program key */
            final String programKey
                    = renderer.getMaterial().getClass().getSimpleName()
                    + renderer.getMesh().getIdentifier();

            /* Generate a program */
            final Program program = renderer.getProgram();
            program.setLayout(entity.getLayout());

            /* Add the generated program to the program list 
            ONLY if the program key doesn't exist already */
            if (!programs.containsKey(programKey)) {
                programs.put(programKey, program);
            }
        }
    }

    /**
     * Check if somethings has been modified in the scene during the frame.
     * @return
     */
    public final synchronized boolean isSomethingsHasBeenTransformed() {
        return somethingsModified && drawCount < 2;
    }

    /**
     * Set that somethings has been modified during the frame.
     */
    public final synchronized void setSomethingsModified() {
        somethingsModified = true;
    }

    /**
     * Change the background image of the scene.
     * @param cubemap
     */
    public final synchronized void setBackground(final TextureCubemap cubemap) {
        this.skyboxCubemap = cubemap;

        for (Entity entity : entities) {
            entity.sendMessage("background", cubemap);
        }

        /* Inform that somethings has been modified and must be draw */
        somethingsModified = true;
    }

    @Override
    public final void onInit(final GL3bc gl) {
        /* Initialize each entity */
        for (Entity entity : entities) {
            entity.init(gl);
        }
    }

    @Override
    public void onInputting(final InputEvent event, final InputStatus status) {
        /* Update each entity */
        for (Entity entity : entities) {
            entity.inputting(event, status);
        }
    }

    @Override
    public final void onUpdate(final GL3bc gl, final float delta) {
        /* Update each entity */
        for (Entity entity : entities) {
            entity.update(gl, delta);
        }

        /* Get if somethings has been modified */
        for (Entity entity : entities) {
            if (entity.isModified()) {
                entity.setModified(false);
                /* Inform that somethings has been modified and must be draw */
                somethingsModified = true;
                drawCount = 0;
                break;
            }
        }
    }

    @Override
    public final void onDisplay(final GL3bc gl) {
        /* Draw only if somethings has been modified */
        if (somethingsModified || drawCount < 2) {
            if (!somethingsModified && drawCount < 2) {
                drawCount ++;
            } else if (somethingsModified) {
                drawCount = 0;
            }

            /* Clear the background color before drawing */
            gl.glClearColor(0.92f, 0.93f, 0.95f, 1.0f);
            gl.glClear(GL3bc.GL_COLOR_BUFFER_BIT);

            /* Update each camera */
            for (Camera camera : cameras) {
                /* Bind the camera framebuffer or the default one */
                camera.bind(gl);

                /* Update camera transformation matrix */
                final Matrix4f viewMatrix = camera.getParent().
                        getTransform().getTransformation().invert();
                final Vector3f cameraPosition = camera.getParent()
                        .getTransform().getWorldPosition();

                /* Update the skybox infos */
                if (skyboxCubemap != null) {
                    final SkyboxMaterial material = new SkyboxMaterial();
                    material.setCubemap(skyboxCubemap);
                    material.setView(viewMatrix);
                    material.setProjection(camera.getProjection());
                    material.setViewPosition(cameraPosition);
                    skyboxProgram.addMaterial(material);
                }

                /* Update each entity with camera infos */
                for (Entity entity : entities) {
                    for (String layout : camera.getRenderedLayouts()) {
                        if (entity.getLayout() == null ? layout == null : entity.getLayout().equals(layout)) {
                            entity.sendMessage("viewMatrix", viewMatrix);
                            entity.sendMessage("projectionMatrix", camera.getProjection());
                            entity.sendMessage("viewPosition", cameraPosition);
                            entity.sendMessage("background", skyboxCubemap);
                            entity.display(gl);
                        }
                    }
                }

                /* Update each program's materials */
                for (Renderer renderer : renderers) {
                    /* Create a program key */
                    final String programKey
                            = renderer.getMaterial().getClass().getSimpleName()
                            + renderer.getMesh().getIdentifier();

                    /* Add the material to the program that have the generated key */
                    if (programs.containsKey(programKey)) {
                        programs.get(programKey).addMaterial(renderer.getMaterial());
                    }
                }

                /* Draw the skybox */
                for (String layout : camera.getRenderedLayouts()) {
                    if (skyboxProgram.getLayout() == null ? layout == null
                            : skyboxProgram.getLayout().equals(layout)) {
                        skyboxProgram.updateProperties();
                        skyboxProgram.bind(gl);
                        skyboxProgram.draw(gl);
                        skyboxProgram.clearMaterials();
                    }
                }

                /* Draw each program and clear the material list */
                programs.values().forEach((program) -> {
                    for (String layout : camera.getRenderedLayouts()) {
                        if (program.getLayout() == null ? layout == null
                                : program.getLayout().equals(layout)) {
                            program.updateProperties();
                            program.bind(gl);
                            program.draw(gl);
                            program.clearMaterials();
                        }
                    }
                });

                /* Draw each camera post process */
                camera.displayPostProcesses(gl);
            }

            /* Pause the drawing loop */
            somethingsModified = false;
        }
    }

    @Override
    public final void onRescale(final GL3bc gl, final int width, final int height) {
        /* Rescale each entity */
        for (Entity entity : entities) {
            entity.rescale(gl, width, height);
        }

        /* Inform that somethings has been modified and must be draw */
        somethingsModified = true;
    }

    @Override
    public final void onDestroy(final GL3bc gl) {
        /* Destroy each entity */
        for (Entity entity : entities) {
            entity.destroy(gl);
        }
    }

    protected abstract void initResources();

    protected abstract void initEntities();

}
