/*
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software. You can use, modify
 * and/ or redistribute the software under the terms of the CeCILL-B license as
 * circulated by CEA, CNRS and INRIA at the following URL "http://www.cecill.info".
 * 
 * As a counterpart to the access to the source code and rights to copy, modify
 * and redistribute granted by the license, users are provided only with a
 * limited warranty and the software's author, the holder of the economic
 * rights, and the successive licensors have only limited liability.
 * 
 * In this respect, the user's attention is drawn to the risks associated with
 * loading, using, modifying and/or developing or reproducing the software by
 * the user in light of its specific status of free software, that may mean that
 * it is complicated to manipulate, and that also therefore means that it is
 * reserved for developers and experienced professionals having in-depth
 * computer knowledge. Users are therefore encouraged to load and test the
 * software's suitability as regards their requirements in conditions enabling
 * the security of their systems and/or data to be ensured and, more generally,
 * to use and operate it in the same conditions as regards security.
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 */
package fr.amap.lidar.amapvox.engine3d.entities.extended;

import com.jogamp.newt.event.InputEvent;
import com.jogamp.opengl.GL3bc;
import fr.amap.commons.util.colors.ColorBars;
import java.awt.Color;
import java.io.IOException;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;
import org.joml.Matrix4f;
import org.joml.Vector3f;
import fr.amap.lidar.amapvox.engine3d.entities.base.Component;
import fr.amap.lidar.amapvox.engine3d.entities.util.VoxelFilter;
import fr.amap.lidar.amapvox.engine3d.entities.util.VoxelSpaceFile;
import fr.amap.lidar.amapvox.engine3d.resources.materials.Material;
import fr.amap.lidar.amapvox.engine3d.resources.materials.StandardMaterial;
import fr.amap.lidar.amapvox.engine3d.resources.meshes.DrawingType;
import fr.amap.lidar.amapvox.engine3d.resources.meshes.Geometry;
import fr.amap.lidar.amapvox.engine3d.resources.meshes.Mesh;
import fr.amap.lidar.amapvox.engine3d.resources.programs.StandardProgram;
import fr.amap.lidar.amapvox.engine3d.resources.textures.Texture2D;
import fr.amap.lidar.amapvox.engine3d.resources.textures.TextureCubemap;
import fr.amap.lidar.amapvox.engine3d.util.InputStatus;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import net.jcip.annotations.GuardedBy;
import net.jcip.annotations.ThreadSafe;
import org.joml.Vector2f;

/**
 * Component used to display a voxel space.
 */
@ThreadSafe
public class VoxelSpace extends Component {

    /**
     * The attached voxel space file.
     */
    @GuardedBy("this")
    private VoxelSpaceFile file;

    /**
     * Filters list (used to hide some voxels).
     */
    private final ConcurrentHashMap<Integer, VoxelFilter> filters;

    /**
     * The size of each voxel drawables.
     */
    @GuardedBy("this")
    private float voxelSize;

    /**
     * If each voxel is displayed has a pixel.
     */
    @GuardedBy("this")
    private boolean drawPixel;

    /**
     * The current colorbar.
     */
    @GuardedBy("this")
    private Color[] colorBar;

    /**
     * The minimum value of the renderer property.
     */
    @GuardedBy("this")
    private float colorMin;

    /**
     * The maximum value of the renderer property.
     */
    @GuardedBy("this")
    private float colorMax;

    /**
     * The default program.
     */
    private final StandardProgram program;

    /**
     * The fast rendering program.
     */
    private final StandardProgram fastProgram;

    /**
     * The materials list.
     */
    private CopyOnWriteArrayList<Material> materials;

    /**
     * The fast rendering materials list.
     */
    private CopyOnWriteArrayList<Material> fastMaterials;

    /**
     * The albedo texture used to display each voxel models.
     */
    @GuardedBy("this")
    private Texture2D albedo;

    /**
     * The current rendered property.
     */
    @GuardedBy("this")
    private String currentProperty;

    /**
     * If the fast drawing mode is enabled or not.
     */
    @GuardedBy("this")
    private boolean drawFastEnabled;

    /**
     * The main camera fov used to resize the pixels.
     */
    @GuardedBy("this")
    private float cameraFOV = Camera.DEFAULT_FOV;

    /**
     * The screen aspect used to resize the voxels displayed has pixels.
     */
    @GuardedBy("this")
    private float screenAspect = 1f;

    /**
     * Create a voxel space component with the Equatorial colorbar.
     * @param voxelMesh
     * @throws IOException
     */
    public VoxelSpace(final Mesh voxelMesh) throws IOException {
        this.filters = new ConcurrentHashMap();
        this.colorBar = ColorBars.getColorBar("Equatorial");
        this.voxelSize = 0.5f;
        this.colorMin = 0f;
        this.colorMax = 1f;
        this.program = new StandardProgram();
        this.albedo = null;
        this.drawFastEnabled = true;
        this.materials = new CopyOnWriteArrayList();
        this.fastMaterials = new CopyOnWriteArrayList();
        program.setMesh(voxelMesh);
        final Mesh fastMesh = new Mesh("FastMesh");
        final Geometry geometry = new Geometry();
        geometry.addVertex(0, 0, 0);
        geometry.addNormal(0, 0, 1);
        geometry.addTexcoord(0, 0);
        geometry.addIndex(0, 0, 0);
        fastMesh.setGeometry(geometry);
        fastMesh.setDrawingType(DrawingType.POINTS);
        fastMesh.setFaceCulling(false);
        this.fastProgram = new StandardProgram();
        fastProgram.setMesh(fastMesh);
    }

    /**
     * Change the voxel space file rendered by this component.
     * @param file
     */
    public final void setFile(final VoxelSpaceFile file) {
        materials.clear();
        fastMaterials.clear();

        synchronized (this) {
            this.file = file;

            for (int i = 0; i < file.getWidth() * file.getHeight() * file.getDepth(); i++) {
                /* Set voxel position by inverting k and j */
                final float x = file.getProperty("i", i) * 2 - file.getWidth() + 1;
                final float y = file.getProperty("k", i) * 2 - file.getDepth() + 1;
                final float z = (file.getHeight() - file.getProperty("j", i)) * 2 - file.getHeight() - 1;

                final Vector3f position = new Vector3f(x, y, z);

                final StandardMaterial material = new StandardMaterial();
                material.setAlbedo(albedo);
                material.setPosition(position);
                materials.add(material);

                final StandardMaterial fastMaterial = new StandardMaterial();
                fastMaterial.setPosition(position);
                fastMaterials.add(fastMaterial);
            }
        }

        setProperty(file.getNames()[0]);
    }

    /**
     * Add or remplace a property filter
     * @param filter
     */
    public final void setFilter(final VoxelFilter filter) {
        if (filter != null) {
            filters.put(filter.getIndex(), filter);
            setProperty(currentProperty);
        }
    }

    /**
     * Remove a filter from the current list.
     * @param index
     */
    public final void removeFilter(final int index) {
        filters.remove(index);
        setProperty(currentProperty);
    }

    /**
     * Get the filter of the specifief property index (if exist).
     * @param index
     * @return
     */
    public final VoxelFilter getFilter(final int index) {
        return filters.get(index);
    }

    /**
     * Get the voxel space file.
     * @return
     */
    public final synchronized VoxelSpaceFile getFile() {
        return file;
    }

    /**
     * Remove each current filter.
     */
    public final void resetFilters() {
        filters.clear();
    }

    /**
     * Set the current displayed property.
     * @param property
     */
    public final void setProperty(final String property) {
        if (property != null && file != null) {
            synchronized (this) {
                currentProperty = property;
            }

            /* Get the min and max value of the new property */
            final float maxValue = file.getMaxPropertyValue(property);
            final float minValue = file.getMinPropertyValue(property);

            /* Get each voxel's value of the new property */
            int voxelCount = 0;

            synchronized (this) {
                voxelCount = file.getWidth() * file.getHeight() * file.getDepth();
            }

            for (int i = 0; i < voxelCount; i++) {
                final float value = file.getProperty(property, i);

                /* Perform a hiding of voxels */
                boolean hiden = false;
                boolean notEqualFilter = false;
                boolean notEqualHiden = false;
                boolean equalHiden = false;
                boolean lesserHiden = false;
                boolean greaterHiden = false;

                /* Make a thread safe copy of the filters list */
                Collection<VoxelFilter> filterList = new ArrayList(0);

                synchronized (this) {
                    filterList = Collections.unmodifiableMap(filters).values();
                }

                /* for each filters */
                for (VoxelFilter next : filterList) {
                    /* Get the filter property value */
                    final float v = file.getProperty(next.getProperty(), i);
                    final float min = file.getMinPropertyValue(next.getProperty());
                    final float max = file.getMaxPropertyValue(next.getProperty());
                    final float n = (v - min) / (max - min);

                    /* Hide the filtered voxels */
                    if (null != next.getFilteringType()) {
                        switch (next.getFilteringType()) {
                            case EQUAL:
                                if (next.isHiden(n)) {
                                    equalHiden = true;
                                }
                                break;
                            case NOTEQUAL:
                                notEqualFilter = true;

                                if (next.isHiden(n)) {
                                    notEqualHiden = true;
                                }
                                break;
                            case LESSER:
                                if (next.isHiden(n)) {
                                    lesserHiden = true;
                                }
                                break;
                            case GREATER:
                                if (next.isHiden(n)) {
                                    greaterHiden = true;
                                }
                                break;
                        }
                    }
                }

                /* Use conditions of each filter state to know 
                if the current voxel is hiden or not. 
                These conditions were found with a chart of Karnaugh*/
                if (notEqualFilter) {
                    hiden = notEqualHiden;
                } else if (equalHiden || (!lesserHiden && greaterHiden)
                        || (lesserHiden && !greaterHiden)) {
                    hiden = true;
                }

                /* Add the voxel property to the list (only if not hiden) */
                final StandardMaterial material
                        = (StandardMaterial) materials.get(i);

                final StandardMaterial fastMaterial
                        = ((StandardMaterial) fastMaterials.get(i));

                if (!Float.isNaN(value) && !hiden) {
                    /* Perform a normalization of the value */
                    final float min = minValue + (maxValue - minValue) * colorMin;
                    final float max = minValue + (maxValue - minValue) * colorMax;
                    float normalizedValue = (value - min) / (max - min);

                    if (normalizedValue < 0f) {
                        normalizedValue = 0f;
                    } else if (normalizedValue > 1f) {
                        normalizedValue = 1f;
                    }

                    /* Add color value from the colorbar */
                    final int colorID
                            = Math.round(normalizedValue * (colorBar.length - 1));
                    material.setMainColor(colorBar[colorID]);
                    material.setHiden(false);
                    fastMaterial.setMainColor(colorBar[colorID]);
                    fastMaterial.setHiden(false);
                } else {
                    material.setHiden(true);
                    fastMaterial.setHiden(true);
                }
            }

            /* Reset and set the materials of the rendering programs */
            program.clearMaterials();
            program.setMaterials(materials);
            fastProgram.clearMaterials();
            fastProgram.setMaterials(fastMaterials);
        }
    }

    /**
     * The min and max of the color bar.
     * @param min
     * @param max
     */
    public final void setColorMinAndMax(final float min, final float max) {
        synchronized (this) {
            this.colorMin = min;
            this.colorMax = max;
        }

        setProperty(currentProperty);
    }
    
    /**
     * Get the min and max values of the current property.
     * @return 
     */
    public final synchronized Vector2f getMinAndMaxPropertyValues() {
        final float minValue = file.getMinPropertyValue(currentProperty);
        final float maxValue = file.getMaxPropertyValue(currentProperty);
        return new Vector2f(minValue + (maxValue - minValue) * colorMin,
                minValue + (maxValue - minValue) * colorMax);
    }

    /**
     * Change if the fast drawing mode is enabled or not.
     * @param drawFastEnabled
     */
    public final synchronized void setDrawFastEnabled(final boolean drawFastEnabled) {
        this.drawFastEnabled = drawFastEnabled;
    }

    /**
     * Set if the pixels are displayed as pixels.
     * @param drawPixel
     */
    public final synchronized void setDrawPixel(final boolean drawPixel) {
        this.drawPixel = drawPixel;
    }

    /**
     * Chnage the current color bar.
     * @param colorBar
     */
    public final void setColorBar(final Color[] colorBar) {
        synchronized (this) {
            this.colorBar = colorBar;
        }
        setProperty(currentProperty);
    }

    /**
     * Get the current color bar.
     * @return
     */
    public final synchronized Color[] getColorBar() {
        return colorBar;
    }

    /**
     * Get the ground height position of the current voxel space file.
     * @return
     */
    public final synchronized float getGroundHeightPosition() {
        if (file != null) {
            return -file.getDepth() - 0.01f;
        }

        return 0f;
    }

    /**
     * Set if the reflection are displayed.
     * @param reflectionEnabled
     */
    public final void setReflectionEnabled(final boolean reflectionEnabled) {
        program.setReflectionEnabled(reflectionEnabled);
        fastProgram.setReflectionEnabled(reflectionEnabled);
        getParent().getScene().setSomethingsModified();
    }

    /**
     * Set the main camera field of view.
     * @param cameraFOV
     */
    public final synchronized void setCameraFOV(float cameraFOV) {
        this.cameraFOV = cameraFOV;
    }

    /**
     * Set the voxel size.
     * @param mode
     */
    public final void setDrawingMode(final float mode) {
        synchronized (this) {
            voxelSize = mode;
        }

        for (Material material : materials) {
            material.setScale(new Vector3f(voxelSize));
        }
        for (Material material : fastMaterials) {
            material.setScale(new Vector3f(voxelSize));
        }

        getParent().getScene().setSomethingsModified();
    }

    /**
     * Set the current albedo texture used to display each voxels.
     * @param albedo
     */
    public final void setAlbedo(final Texture2D albedo) {
        synchronized (this) {
            this.albedo = albedo;
        }

        for (Material material : materials) {
            ((StandardMaterial) material).setAlbedo(albedo);
        }
    }

    /**
     * Get the camera Z position to display the entire current voxel space file
     * datas.
     * @param fov
     * @return
     */
    public final float getCameraRetreat(final float fov) {
        if (file != null) {
            return (float) ((file.getWidth() + file.getHeight() + file.getDepth())
                    / Math.sin(Math.toRadians(fov) * 2));
        }

        return 0f;
    }

    @Override
    public final void onInit(final GL3bc gl) {
        /* Nothing needed to do */
    }

    @Override
    public final void onInputting(final InputEvent event, final InputStatus status) {
        /* Nothing needed to do */
    }

    @Override
    public final void onUpdate(final GL3bc gl, final float delta) {
        /* Nothing needed to do */
    }

    @Override
    public final void onDisplay(final GL3bc gl) {
        if (((getParent().getScene().isSomethingsHasBeenTransformed()
                && drawFastEnabled) || drawPixel) && fastProgram != null) {
            /* Get each pixel size aspects */
            float voxelSizeAspect = ((file.getWidth() + file.getHeight() + file.getDepth()) / 3);
            voxelSizeAspect = voxelSizeAspect / (voxelSizeAspect + 1);
            float cameraFovAspect = (1 - cameraFOV / (cameraFOV + 1));
            float drawFastAspect = drawPixel ? 1f : 0.9f;

            /* Change the size of the pixels */
            gl.glPointSize(voxelSize * (1 - voxelSizeAspect)
                    * cameraFovAspect * drawFastAspect * screenAspect * 20);

            /* Draw the fast rendering mode program */
            fastProgram.updateProperties();
            fastProgram.bind(gl);
            fastProgram.draw(gl);
        } else if (program != null) {
            /* Draw the default program */
            program.updateProperties();
            program.bind(gl);
            program.draw(gl);
        }
    }

    @Override
    public final void onRescale(final GL3bc gl, final int width, final int height) {
        /* Update the screen aspect used to resize each pixels */
        screenAspect = (float) (width + height);
    }

    @Override
    public final void onDestroy(final GL3bc gl) {
        /* Nothing needed to do */
    }

    @Override
    protected final synchronized void receiveMessage(final String name, final Object object) {
        if (null != name) {
            switch (name) {
                case "viewMatrix":
                    for (Material material : materials) {
                        ((StandardMaterial) material).setView((Matrix4f) object);
                    }
                    for (Material material : fastMaterials) {
                        ((StandardMaterial) material).setView((Matrix4f) object);
                    }
                    break;
                case "projectionMatrix":
                    for (Material material : materials) {
                        ((StandardMaterial) material).setProjection((Matrix4f) object);
                    }
                    for (Material material : fastMaterials) {
                        ((StandardMaterial) material).setProjection((Matrix4f) object);
                    }
                    break;
                case "viewPosition":
                    for (Material material : materials) {
                        ((StandardMaterial) material).setViewPosition((Vector3f) object);
                    }
                    break;
                case "background":
                    for (Material material : materials) {
                        ((StandardMaterial) material).setCubemap((TextureCubemap) object);
                    }
                    break;
            }
        }
    }

}
